#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef linux
#include <signal.h>
#include <errno.h>
#include <pty.h>
#include <utmp.h>
#else
#include <libutil.h>
#endif
#include "script.h"

static void
ssh(const char *host, const char *command, const char *revision)
{
    execlp("ssh", "ssh", "-q", "-t", "-F", "ssh_config",
	   host, command, revision, (char *)NULL);
}

#ifdef linux
static int theMaster;

static void
childKilled(int signalNumber UNUSED)
{
    int status;

    while (wait3(&status, WNOHANG, 0) > 0) {
        continue;
    }
    close(theMaster);
}
#endif

static int
scriptWithPty(FILE *fp, int master, int slave,
	      const char *host, const char *command, const char *revision)
{
    static const char SIGN_STR[] = "\r\nOK\r\n";
    static const size_t SIGN_LEN = sizeof(SIGN_STR) - 1;
    int pid, status;
    char buf[BUFSIZ];
    ssize_t size;
#ifdef linux
    sig_t orig;
#endif

    if ((pid = fork()) < 0) {
	return 0;
    }
    if (pid == 0) {
	(void)close(master);
	(void)fclose(fp);
        login_tty(slave);
	ssh(host, command, revision);
	_exit(0);
    }
#ifdef linux
    theMaster = master;
    orig = signal(SIGCHLD, childKilled);
    if (kill(pid, 0) < 0 && errno == ESRCH) {
        childKilled(SIGCHLD);
    }
#endif
    while ((size = read(master, buf, sizeof(buf))) > 0) {
	(void)fwrite(buf, 1, size, fp);
    }
    (void)fflush(fp);
    while (wait3(&status, WNOHANG, 0) > 0) {
	continue;
    }
#ifdef linux
    signal(SIGCHLD, orig);
#endif
    if ((size = ftell(fp)) >= 0
	&& (size_t)size >= SIGN_LEN
	&& fseek(fp, -SIGN_LEN, SEEK_END) >= 0
	&& fread(buf, 1, SIGN_LEN, fp) == SIGN_LEN
	&& memcmp(buf, SIGN_STR, SIGN_LEN) == 0) {
	status = 1;
    }
    else {
	status = 2;
    }
    return status;
}

static int
scriptWithFile(FILE *fp, const char *host, const char *command,
	       const char *revision)
{
    int master, slave, status;

    if (openpty(&master, &slave, NULL, NULL, NULL) == -1)
	return 0;
    status = scriptWithPty(fp, master, slave, host, command, revision);
    (void)close(master);
    (void)close(slave);
    return status;
}

int
script(const char *path, const char *host, const char *command,
       const char *revision)
{
    FILE *fp;
    int status;

    if ((fp = fopen(path, "w+")) == NULL)
	return 0;
    status = scriptWithFile(fp, host, command, revision);
    (void)fclose(fp);
    return status;
}
